/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <ctype.h>
#include <malloc.h>
#include <gtk/gtk.h>

#include "auxil.h"

/* block size as used in readline */

#define _BLOCKSIZE 256

/* global functions */

/* Copies src to *dest upto and including the trailing zero after
 * reallocating memory for *dest, which should either be NULL, or a
 * pointer returned by a previous malloc()/realloc(). If src is NULL,
 * *dest is also set to NULL, after free()ing any memory poited by it.
 * Returns 0 on success. If realloc() fails, the return value is nonzero,
 * and *dest is untouched. */
int
string_copy(char *src, char **dest)
{
  char *tmp_dest;

  if (src == NULL) {
    free(*dest);
    (*dest) = NULL;
  }
  else {
    tmp_dest = realloc((*dest),  strlen(src) + 1);
    if (tmp_dest == NULL) return 1;
    (*dest) = tmp_dest;
    strcpy(*dest, src);
  }
  return 0;
}

/* Reads a line of text from stream using readline(), and sets *ptr1
 * and *ptr2 so that,
 * 1. If the line has all space characters both are set to NULL.
 * 2. If there is only one word, a '\0' character is added to the
 *    end of it, *ptr1 is set to point to the first character of
 *    that word, and *ptr2 is set to NULL.
 * 3. If there is more than one word, the first word is zero terminated
 *    as in 2 above, and another zero is added just after the last word.
 *    *ptr1 is set to point to the first word, and *ptr2 to to the second
 *    word.
 * Returns a pointer to the whole string, which can be passed to free().
 * If reading from stream fails, *ptr1 and *ptr2 are set to NULL and
 * the functions returns NULL. */
char*
read_two_strings(char **ptr1, char **ptr2, FILE *stream)
{
  char *linebuf = NULL;
  int i, i0;

  if ((linebuf = readline(stream)) == NULL) {
    /* reading from stream failed */
    (*ptr1) = NULL;
    (*ptr2) = NULL;
    return NULL;
  }

  /* locate the first non white space character */
  i = 0;
  while (isspace(linebuf[i]) && (linebuf[i] != '\n'))
    i++;

  if (linebuf[i] == '\n'){
    /* blank line */
    (*ptr1) = NULL;
    (*ptr2) = NULL;
    return linebuf;
  }

  /* *ptr1 points to the first word */
  (*ptr1) = linebuf + i;

  /* locate the end of the first word can't zero terminate here,
     because the next character may be '\n' */
  while (!isspace(linebuf[i])) i++;
  i0 = i;

  /* locate the next word terminate the first word */
  while (isspace(linebuf[i]) && (linebuf[i] != '\n'))
    i++;

  if (linebuf[i] == '\n'){
    /* only one word available */
    linebuf[i0] = 0;
    (*ptr2) = NULL;
    return linebuf;
  }

  /* zero terminate the first word */
  linebuf[i0] = 0;

  /* ptr2 points to the rest */
  (*ptr2) = linebuf + i;

  /* locate the end of line */
  while (linebuf[i] != '\n')
    i++;

  /* find the end of the last word and zero terminate it */
  while (isspace(linebuf[i]))
    i--;
  linebuf[i + 1] = 0;

  return(linebuf);
}

/* Capitalizes a null terminated string. */
void
capitalize(char *s)
{
  if (s == NULL) return;
  while (*s){
    (*s) = toupper(*s);
    s++;
  }
}

/* Reads a string of characters from stream unto and including '\n',
 * into a block of dynamically allocated memory and returns a pointer to
 * it. The size of this block is a multiple of _BLOCKSIZE. Returns NULL if
 * realloc() fails.
 * Note : BDF specification says that lines are either terminated by
 * character '\015' or '\012'. However, this function tests the end
 * of line with '\n' and not with the character values. */
char
*readline(FILE *stream)
{
  char *linebuf = NULL, *tmpbuf;
  int ptr = 0, bufsize = 0, c = 0;

  while (c != '\n'){
    c = getc(stream); 

    if (c == EOF){
      if (ptr)
	/* last line ended with EOF instead of '\n' */
	c = '\n';
      else
	/* real end of file */
	return NULL;
    }

    if (ptr == bufsize){
      /* linebuf is full */
      bufsize += _BLOCKSIZE;
      tmpbuf = realloc(linebuf, bufsize);
      if (tmpbuf == NULL){
	free(linebuf);
	return NULL;
      }
      linebuf = tmpbuf;
    }

    linebuf[ptr] = c;
    ptr++;
  }

  tmpbuf = realloc(linebuf, ptr);
  if (tmpbuf == NULL){
    free(linebuf);
    return NULL;
  }
 
  return(tmpbuf);
}

