/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>

#include <ctype.h>
#include <malloc.h>

#include "auxil.h"
#include "bdf.h"
#include "bitmap.h"
#include "raster_font.h"
#include "raster_glyph.h"

/* local function declarations */

static GfeRasterGlyph *bdf_read_next_glyph_from_file(GfeRasterFont *font,
		                                     FILE *bdf_file);

/* global functions */

/* Reads the file `filename' in BDF format into a GfeRasterFont structure
 * and returns a pointer to it. If there is an error during reading the
 * header NULL is returned. If at least one glyph is read from the file,
 * it is considered as a valid file and a non-NULL pointer is returned. */
GfeRasterFont
*bdf_font_read_from_file(char *filename)
{
  GfeRasterFont *font;
  GfeRasterGlyph *glyph;
  FILE *bdf_file;
  char *linebuf;
  char *ptr1;
  char *ptr2;
  gboolean header = TRUE;
  int i;
  int num_properties;
  int num_glyphs;

  if ((bdf_file = fopen(filename, "rt")) == NULL)
    return NULL;

  font = gfe_raster_font_new();          
  gfe_font_set_filename(GFE_FONT(font), filename);
  GFE_FONT(font)->format = FONT_FORMAT_BDF21;

  do {
    linebuf = read_two_strings(&ptr1, &ptr2, bdf_file);
    capitalize(ptr1);
    if (linebuf == NULL){
      gfe_raster_font_destroy(font);
      return NULL;
    }

    if (ptr1 == NULL){
      /* blank line */
      ;
    }
    else if (strcmp(ptr1, "STARTFONT") == 0){
      if (strcmp(ptr2, "2.2") == 0)
        GFE_FONT(font)->format = FONT_FORMAT_BDF22;
    }
    else if (strcmp(ptr1, "CONTENTVERSION") == 0){
      gfe_font_set_content_version(GFE_FONT(font), ptr2);
    }
    else if (strcmp(ptr1, "FONT") == 0){
      gfe_font_set_name(GFE_FONT(font), ptr2);
    }
    else if (strcmp(ptr1, "SIZE") == 0){
      sscanf(ptr2, "%d %d %d", &(font->point_size), 
		      &(font->x_resolution),
		      &(font->y_resolution));
    }
    else if (strcmp(ptr1, "FONTBOUNDINGBOX") == 0){
      sscanf(ptr2, "%d %d %d %d", &(font->bbx_width),
		      &(font->bbx_height),
		      &(font->bbx_x_offset),
		      &(font->bbx_y_offset));
    }
    else if (strcmp(ptr1, "METRICSSET") == 0){
      sscanf(ptr2, "%d", &(font->metricsset));
    }
    else if (strcmp(ptr1, "SWIDTH") == 0){
      sscanf(ptr2, "%f %f", &(font->sw_x), &(font->sw_y));
    }
    else if (strcmp(ptr1, "DWIDTH") == 0){
      sscanf(ptr2, "%d %d", &(font->dw_x), &(font->dw_y));
    }
    else if (strcmp(ptr1, "SWIDTH1") == 0){
      sscanf(ptr2, "%f %f", &(font->sw1_x), &(font->sw1_y));
    }
    else if (strcmp(ptr1, "DWIDTH1") == 0){
      sscanf(ptr2, "%d %d", &(font->dw1_x), &(font->dw1_y));
    }
    else if (strcmp(ptr1, "VVECTOR") == 0){
      sscanf(ptr2, "%d %d", &(font->vv_x), &(font->vv_y));
    }
    else if (strcmp(ptr1, "STARTPROPERTIES") == 0){
      sscanf(ptr2, "%d", &(font->num_properties));
      num_properties = font->num_properties;
      font->property = malloc(sizeof(char*) * num_properties);
      font->value = malloc(sizeof(char*) * num_properties);

      for (i = 0; i < num_properties; i++){
	font->property[i] = NULL;
	font->value[i] = NULL;
      }

      for (i = 0; i < num_properties; i++){
	free(linebuf);
        linebuf = read_two_strings(&ptr1, &ptr2, bdf_file);
        if (linebuf == NULL){
	  gfe_raster_font_destroy(font);
          return NULL;
        }

	string_copy(ptr1, font->property + i);
	string_copy(ptr2, font->value + i);
      }

      /* read the line ENDPROPERTIES */
      free(linebuf);
      linebuf = read_two_strings(&ptr1, &ptr2, bdf_file);
      capitalize(ptr1);
    }
    else if (strcmp(ptr1, "CHARS") == 0){
      /* beginning of individual glyphs, end of header */
      /* don't manipulate font->num_glyphs directly, this allows
       * to recoved something from damaged font files */
      sscanf(ptr2, "%d", &num_glyphs);
      header = FALSE;
    }

    free(linebuf);

  } while (header);

  while ((glyph = bdf_read_next_glyph_from_file(font, bdf_file))) {
    /* read individual glyphs */
    gfe_font_add_glyph(GFE_FONT(font), GFE_GLYPH(glyph));
  }

  if (GFE_FONT(font)->first_glyph == NULL){
    /* no glyphs have been read */
    gfe_raster_font_destroy(font);
    return NULL;
  }

  fclose(bdf_file);
  return (font);
}

/* Save font in BDF format. The completeness of font is not checked.
 * Always use version 2.1 BDf, therefore global values of swidth etc.
 * are not written. This should be changed in a future versions when
 * font format is also determined from flags. Returns zero on success,
 * nonzero if font is NULL, font->filename is NULL or file error.
 * Error checking is NOT yet added to individual fprintf()s. */
int
bdf_font_save(GfeRasterFont *font)
{
  FILE *bdf_file;
  GfeRasterGlyph *glyph;
  int i, j, width, n;

  if (font == NULL)
    return 1;
  if (GFE_FONT(font)->filename == NULL)
    return 1;
  if ((bdf_file = fopen(GFE_FONT(font)->filename, "wt")) == NULL)
    return 1;

  fprintf(bdf_file, "STARTFONT ");
  fprintf(bdf_file, "2.1\n");

  fprintf(bdf_file, "COMMENT Font saved by GNU Font Editor\n");

  if (GFE_FONT(font)->format == FONT_FORMAT_BDF22){
    if (GFE_FONT(font)->content_version){
      fprintf(bdf_file, "CONTENTVERSION %s\n", GFE_FONT(font)->content_version);
    }
  }

  fprintf(bdf_file, "FONT %s\n", GFE_FONT(font)->name);
  fprintf(bdf_file, "SIZE %d %d %d\n", font->point_size,
		  font->x_resolution, font->y_resolution);
  if (font->bbx_width != GFE_UNDEFINED){
    fprintf(bdf_file, "FONTBOUNDINGBOX %d %d %d %d\n",
		    font->bbx_width, font->bbx_height,
		    font->bbx_x_offset, font->bbx_y_offset);
  }

  if (GFE_FONT(font)->format == FONT_FORMAT_BDF22){
    if (font->metricsset != GFE_UNDEFINED){
      fprintf(bdf_file, "METRICSSET %d\n", font->metricsset);
    }
  }
  if (font->sw_x != GFE_UNDEFINED){
    fprintf(bdf_file, "SWIDTH %.1f %.1f\n", font->sw_x, font->sw_y);
  }
  if (font->dw_x != GFE_UNDEFINED){
    fprintf(bdf_file, "DWIDTH %d %d\n", font->dw_x, font->dw_y);
  }
  if (GFE_FONT(font)->format == FONT_FORMAT_BDF22){
    if (font->sw1_x != GFE_UNDEFINED){
      fprintf(bdf_file, "SWIDTH1 %.1f %.1f\n", font->sw1_x, font->sw1_y);
    }
    if (font->dw1_x != GFE_UNDEFINED){
      fprintf(bdf_file, "DWIDTH1 %d %d\n", font->dw1_x, font->dw1_y);
    }
    if (font->vv_x != GFE_UNDEFINED){
      fprintf(bdf_file, "VVECTOR %d %d\n", font->vv_x, font->vv_y);
    }
  }

  if (font->num_properties > 0){
    fprintf(bdf_file, "STARTPROPERTIES %d\n", font->num_properties);
    for (i = 0; i < font->num_properties; i++){
      fprintf(bdf_file, "%s %s\n", font->property[i], font->value[i]);
    }
    fprintf(bdf_file, "ENDPROPERTIES\n");
  }

  fprintf(bdf_file, "CHARS %d\n", GFE_FONT(font)->num_glyphs);

  glyph = GFE_RASTER_GLYPH(GFE_FONT(font)->first_glyph);

  while (glyph){
    fprintf(bdf_file, "STARTCHAR ");
    if (GFE_GLYPH(glyph)->name){
      fprintf(bdf_file, "%s\n", GFE_GLYPH(glyph)->name);
    }
    else{
      fprintf(bdf_file, "%X\n", GFE_GLYPH(glyph)->encoding);
    }
    if (GFE_GLYPH(glyph)->encoding != GFE_GLYPH_NO_ENCODING){
      fprintf(bdf_file, "ENCODING %d\n", GFE_GLYPH(glyph)->encoding);
    }
    if (glyph->sw_x != GFE_UNDEFINED){
      fprintf(bdf_file, "SWIDTH %.1f %.1f\n", glyph->sw_x, glyph->sw_y);
    }
    if (glyph->dw_x != GFE_UNDEFINED){
      fprintf(bdf_file, "DWIDTH %d %d\n", glyph->dw_x, glyph->dw_y);
    }

    if (GFE_FONT(font)->format == FONT_FORMAT_BDF22){
      if (glyph->sw1_x != GFE_UNDEFINED){
        fprintf(bdf_file, "SWIDTH1 %.1f %.1f\n", glyph->sw1_x, glyph->sw1_y);
      }
      if (glyph->dw1_x != GFE_UNDEFINED){
        fprintf(bdf_file, "DWIDTH1 %d %d\n", glyph->dw1_x, glyph->dw1_y);
      }
      if (glyph->vv_x != GFE_UNDEFINED){
        fprintf(bdf_file, "VVECTOR %d %d\n", glyph->vv_x, glyph->vv_y);
      }
    }

    if (glyph->bitmap->height != GFE_UNDEFINED){
      fprintf(bdf_file, "BBX %d %d %d %d\n",
		      glyph->bbx_width, glyph->bbx_height,
		      glyph->bbx_x_offset, glyph->bbx_y_offset);
    }
    fprintf(bdf_file, "BITMAP\n");
    for (i = 0; i < glyph->bbx_height; i++){
      width = (glyph->bbx_width - 1) / 8 + 1;
      for (j = 0; j < width; j++){
	n = glyph->bitmap->data[i][j];
	fprintf(bdf_file, "%02X", n);
      }
      fprintf(bdf_file, "\n");
    }
    fprintf(bdf_file, "ENDCHAR\n");

    glyph = GFE_RASTER_GLYPH(GFE_GLYPH(glyph)->next);
  }

  fprintf(bdf_file, "ENDFONT\n");

  fclose(bdf_file);
  return 0;
}

/* local functions */

/* Reads the next glyph from bdf_file in BDF format to a GfeRasterGlyph 
 * structure based on `font' and returns a pointer. Returns NULL in
 * case of a failure or EOF. */
static
GfeRasterGlyph *bdf_read_next_glyph_from_file(GfeRasterFont *font,
		                              FILE *bdf_file)
{
  GfeRasterGlyph *glyph;
  char *linebuf = NULL;
  char *ptr1, *ptr2;
  char hex_string[3];
  int done = FALSE;
  int header = TRUE;
  int i, j, n = 0;
  int rows, columns, row, column;

  glyph = gfe_raster_glyph_new(font);

  do {
    linebuf = read_two_strings(&ptr1, &ptr2, bdf_file);
    if (linebuf == NULL){
      gfe_raster_glyph_destroy(glyph);
      return NULL;
    }
    capitalize(ptr1);

    if (strcmp(ptr1, "STARTCHAR") == 0){
      string_copy(ptr2, &(GFE_GLYPH(glyph)->name));
    }
    else if (strcmp(ptr1, "ENCODING") == 0){
      sscanf(ptr2, "%d", &(GFE_GLYPH(glyph)->encoding));
    }
    else if (strcmp(ptr1, "SWIDTH") == 0){
      sscanf(ptr2, "%f %f", &(glyph->sw_x), &(glyph->sw_y));
    }
    else if (strcmp(ptr1, "DWIDTH") == 0){
      sscanf(ptr2, "%d %d", &(glyph->dw_x), &(glyph->dw_y));
    }
    else if (strcmp(ptr1, "SWIDTH1") == 0){
      sscanf(ptr2, "%f %f", &(glyph->sw1_x), &(glyph->sw1_y));
    }
    else if (strcmp(ptr1, "DWIDTH1") == 0){
      sscanf(ptr2, "%d %d", &(glyph->dw1_x), &(glyph->dw1_y));
    }
    else if (strcmp(ptr1, "VVECTOR") == 0){
      sscanf(ptr2, "%d %d", &(glyph->vv_x), &(glyph->vv_y));
    }
    else if (strcmp(ptr1, "BBX") == 0){
      sscanf(ptr2, "%d %d %d %d", &(glyph->bbx_width),
		      &(glyph->bbx_height),
		      &(glyph->bbx_x_offset),
		      &(glyph->bbx_y_offset));
    }
    else if (strcmp(ptr1, "BITMAP") == 0){
      /* beginning of bitmap data, end of header */
      header = FALSE;
    }

    free(linebuf);

  } while (header);

  gfe_bitmap_allocate(glyph->bitmap, glyph->bbx_width, glyph->bbx_height);
  rows = glyph->bbx_height;
  columns = (glyph->bbx_width - 1) / 8 + 1;
  row = 0;

  for (i = 0; i < rows; i++){
    for (j = 0; j < columns; j++){
      glyph->bitmap->data[i][j] = 0;
    }
  }

  hex_string[2] = 0;
  do {
    linebuf = read_two_strings(&ptr1, &ptr2, bdf_file);
    if (linebuf == NULL){
      gfe_raster_glyph_destroy(glyph);
      return NULL;
    }

    capitalize(ptr1);

    if (strcmp(ptr1, "ENDCHAR") == 0){
      /* end of glyph */
      done=TRUE;
    }
    else {
      i = 0;
      column = 0;
      while (isxdigit(ptr1[i])){
	hex_string[0] = ptr1[i++];
	hex_string[1] = ptr1[i++];
	sscanf(hex_string, "%X", &n);
	if ((column < columns) && (row < rows))
	  glyph->bitmap->data[row][column] = n;
	column++;
      }
      row++;
    }

    free(linebuf);

  } while (!done);

  return (glyph);
}

