/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>
#include <malloc.h>

#include "bitmap.h"

/* local variables */

/* to handle individual bits */
static guint8 bit[] = {128, 64, 32, 16, 8, 4, 2, 1};

/* global functions */

/* Create a GfeBitmap structure and returns a pointer to it.
 * See structures.txt */
GfeBitmap*
gfe_bitmap_new(void)
{
  GfeBitmap *bitmap;

  bitmap = malloc(sizeof(GfeBitmap));

  bitmap->width = 0;
  bitmap->height = 0;
  bitmap->data = NULL;

  return (bitmap);
}

/* Allocates space in bitmap for a 1 bit bitmap of size widthxheight
 * and sets all its bits to zero. Free any previously allocated space, 
 * if any. Returns zero on success, nonzero if malloc fails. */
int
gfe_bitmap_allocate(GfeBitmap *bitmap, int width, int height)
{
  int i, hsize;

  gfe_bitmap_free(bitmap);

  bitmap->data = malloc(sizeof(guint8*) * height);

  if (bitmap->data == NULL) return 1;

  bitmap->width = width;
  bitmap->height = height;

  hsize = (width - 1) / 8 + 1;
  for (i = 0; i < height; i++){
    bitmap->data[i] = malloc(sizeof(guint8) * hsize);
    if (bitmap->data[i] == NULL) {
      gfe_bitmap_free(bitmap);
      return 1;
    }
  }

  gfe_bitmap_clear(bitmap);

  return 0;
}

/* Sets all the bits of bitmap to zero. */
void
gfe_bitmap_clear(GfeBitmap *bitmap)
{
  int i, j, hsize;

  if (bitmap->data == NULL) return;

  hsize = (bitmap->width - 1) / 8 + 1;
  for (i = 0; i < bitmap->height; i++){
    for (j = 0; j < hsize; j++){
      bitmap->data[i][j] = 0;
    }
  }
}

/* Free dynamically allocated memory of bitmap. See structures.txt.
 * Also sets the size to 0x0 */
void
gfe_bitmap_free(GfeBitmap *bitmap)
{
  int i;

  if (bitmap->data == NULL) return;

  for (i = 0; i < bitmap->height; i++){
    free(bitmap->data[i]);
  }
  free(bitmap->data);
  bitmap->data = NULL;

  bitmap->width = 0;
  bitmap->height = 0;
}

/* Destroys bitmap. See structures.txt */
void
gfe_bitmap_destroy(GfeBitmap *bitmap)
{
  gfe_bitmap_free(bitmap);
  free(bitmap);
}

/* Turns on the (column, row)th BIT of bitmap, if it's within
 * the bitmap's boundary. */
void
gfe_bitmap_set_bit(GfeBitmap *bitmap, int column, int row)
{
  if ((column > bitmap->width) || (row > bitmap->height)
      || (column < 0) || (row < 0)) return;
  bitmap->data[row][column / 8] |= bit[column % 8];
}

/* Sets the (column, row)th **BYTE** of bitmap to value, it it's
 * within bitmap's boundary. */
void
gfe_bitmap_set_byte(GfeBitmap *bitmap, int column, int row, guint8 value)
{
  if ((column > (bitmap->width - 1) / 8) || (row > bitmap->height)
      || (column < 0) || (row < 0)) return;
  bitmap->data[row][column] = value;
}

/* Turns off the (column, row)th BIT of bitmap, if it's within
 * the bitmap's boundary. */
void
gfe_bitmap_unset_bit(GfeBitmap *bitmap, int column, int row)
{
  if ((column > bitmap->width) || (row > bitmap->height)
      || (column < 0) || (row < 0)) return;
  if (bitmap->data[row][column / 8] & bit[column % 8])
    bitmap->data[row][column / 8] -= bit[column % 8];
}

/* Returns TRUE if (column, row)th bit of bitmap is on, FALSE if not.
 * IF (column, row) is outside the bitmap, returns FALSE */
gboolean
gfe_bitmap_get_bit(GfeBitmap *bitmap, int column, int row)
{
  if ((column > bitmap->width) || (row > bitmap->height)
      || (column < 0) || (row < 0))
    return FALSE;
  if (bitmap->data[row][column / 8] & bit[column % 8])
    return TRUE;
  else
    return FALSE;
}

