/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>

#include "bitmap.h"
#include "editor.h"
#include "font.h"
#include "glyph.h"
#include "glyph_toolbar.h"
#include "gui.h"
#include "raster_font.h"
#include "raster_glyph.h"

/* macro(s) */

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif

/* global variables */

/* The open font list is a two way linked list. Three pointers are used.
 * All of them are set to NULL, if no font files are open. They are of
 * generic GfeFont type, but the actual variables will always be a type
 * inherited from it. */
GfeFont *first_font = NULL;
GfeFont *last_font = NULL;
GfeFont *current_font = NULL;

/* The currentr_glyph of the current_glyph is copied to a seperate
 * structure, and editing is done on it. */
GfeGlyph *current_glyph = NULL;

/* Flag to indicate busy state. A drawing function keeps this set to
 * TRUE while working. Also these functions returns immediately if
 * working is set to TRUE avoiding multiple instances. */
gboolean working = FALSE;

/* Flag to indicate drawing state. If this is FALSE, erasing is done
 * instead of drawing. This is only until other sophisticated means of 
 * drawing is introduced. */
gboolean drawing = TRUE;

/* Flag to indicate mouse pressed state. Used by drawing functions. */
static gboolean pressed = FALSE;

/* Size of the raster, currently fixed. */
static int boxsize = 10;

/* The main drawing area of GFE. */
static GtkWidget *drawing_area = NULL;

/* Backing up pixmap of drawing_area. */
static GdkPixmap *pixmap = NULL;

/* Width of current pixmap. */
static int pixmap_width = 0;

/* Height of current pixmap. */
static int pixmap_height = 0;

/* local function declarations */

static void draw_glyph(void);
static void color_box(int col, int row, int black);
static gint expose_event(GtkWidget *widget, GdkEventExpose *event);
static gint button_press_event (GtkWidget *widget, GdkEventButton *event);
static gint button_release_event (GtkWidget *widget, GdkEventButton *event);
static gint motion_notify_event (GtkWidget *widget, GdkEventButton *event);

/* global functions */

/* Creates the main drawing area, sets event handlers and returns a
 * pointer to it. All the child widgets are shown, but not the
 * drawing area itself. */
GtkWidget
*create_drawing_area(void)
{
  drawing_area = gtk_drawing_area_new();  

  gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 1, 1);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "expose_event",
		     GTK_SIGNAL_FUNC(expose_event), NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "button_press_event",
		     GTK_SIGNAL_FUNC(button_press_event), NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "button_release_event",
		     GTK_SIGNAL_FUNC(button_release_event), NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "motion_notify_event",
		     GTK_SIGNAL_FUNC(motion_notify_event), NULL);

  gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
                              | GDK_BUTTON_PRESS_MASK
                              | GDK_BUTTON_RELEASE_MASK
                              | GDK_BUTTON1_MOTION_MASK);
  return (drawing_area);
}

/* This function should be called **BEFORE** changing the current_glyph
 * of current_font. Copies contents of current_glyph to
 * current_font->current_glyph after suitable changes, such as prooning
 * for raster glyphs. */
int
set_current_glyph_pre()
{
  if (current_glyph == NULL) return 0;

  switch (GFE_GLYPH_TYPE(current_glyph)) {
    case FONT_TYPE_RASTER:
      gfe_raster_glyph_copy_to_font(GFE_RASTER_GLYPH(current_glyph),
		                    GFE_RASTER_FONT(current_font));
      break;
  }
  return 0;
}
 
/* This function should be called **AFTER** changing the current_glyph
 * of the current font. Copies the contents of current_font->current_glyph
 * to current_glyph after suitable changes, such as expanding the BBX for
 * raster glyphs. Also calls draw_glyph to redraw the screen. If
 * current_glyph is NULL, drawing_area is cleared. */
int
set_current_glyph_post()
{
  if ((current_font == NULL) || (current_font->current_glyph == NULL)) {
    if (current_glyph){
      switch (GFE_GLYPH_TYPE(current_glyph)) {
        case FONT_TYPE_RASTER:
	  gfe_raster_glyph_destroy(GFE_RASTER_GLYPH(current_glyph));
	  current_glyph = NULL;
	  break;
      }
    }
    draw_glyph();
    return 0;
  }

  switch (GFE_FONT_TYPE(current_font)) {
    case FONT_TYPE_RASTER:
      if (current_glyph == NULL){
        current_glyph = GFE_GLYPH(
			gfe_raster_glyph_new(GFE_RASTER_FONT(current_font)));
	/* change these 'just' numbers to meaningful ones */
        gfe_bitmap_allocate(GFE_RASTER_GLYPH(current_glyph)->bitmap, 30, 30);
        GFE_RASTER_GLYPH(current_glyph)->bbx_width = 30;
        GFE_RASTER_GLYPH(current_glyph)->bbx_height = 30;
        GFE_RASTER_GLYPH(current_glyph)->bbx_x_offset = -10;
        GFE_RASTER_GLYPH(current_glyph)->bbx_y_offset = -10;
      }
      gfe_raster_glyph_copy_from_font(GFE_RASTER_FONT(current_font),
		                      GFE_RASTER_GLYPH(current_glyph));
      break;
  }

  draw_glyph();
  glyph_toolbar_set();

  return 0;
}

/* local functions */

/* Sets the colour of a box in the editing window. If black is TRUE,
 * it is coloured in black, otherwise in white. */
static void
color_box(int col, int row, int black)
{
  GdkRectangle update_rect;

  if (pixmap == NULL) return; 

  update_rect.x = col * (boxsize + 1) + 1;
  update_rect.y = row * (boxsize + 1) + 1;
  update_rect.width = boxsize;
  update_rect.height = boxsize;

  if (black){
    gdk_draw_rectangle(pixmap, drawing_area->style->black_gc, TRUE,
                   update_rect.x, update_rect.y,
                   update_rect.width, update_rect.height);
  }
  else {
    gdk_draw_rectangle(pixmap, drawing_area->style->white_gc, TRUE,
                   update_rect.x, update_rect.y,
                   update_rect.width, update_rect.height);
  }

  gtk_widget_draw(drawing_area, &update_rect);

  sm_color_box(col, row, black);

  while (gtk_events_pending())
    gtk_main_iteration();
}

/* Exposet_event handler of drawing_area. Just copies from pixmap to
 * drawing_area. */
static gint
expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  if (pixmap == NULL) return FALSE;
  gdk_draw_pixmap(widget->window,
                  widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
	          pixmap, event->area.x, event->area.y,
                  event->area.x, event->area.y,
                  event->area.width, event->area.height);
  return FALSE;
}

/* Button_press event handler of drawing_area. Begins drawing/erasing. */
static gint
button_press_event(GtkWidget *widget, GdkEventButton *event)
{
  int row, column;

  if (working) return TRUE;
  if (current_glyph == NULL) return TRUE;
  if ((event->x >= pixmap_width) || (event->y >= pixmap_height)) return TRUE;
  working = TRUE;

  row = event->y / (boxsize + 1);
  column = event->x / (boxsize + 1);

  if (event->button == 1){
    pressed = TRUE;
    switch (GFE_FONT_TYPE(current_font)) {
      case FONT_TYPE_RASTER:
	if (gfe_bitmap_get_bit(GFE_RASTER_GLYPH(current_glyph)->bitmap,
			       column, row)){
	  drawing = FALSE;
	  gfe_bitmap_unset_bit(GFE_RASTER_GLYPH(current_glyph)->bitmap,
			       column, row);
	  color_box(column, row, FALSE);
	}
	else {
	  drawing = TRUE;
	  gfe_bitmap_set_bit(GFE_RASTER_GLYPH(current_glyph)->bitmap,
			     column, row);
	  color_box(column, row, TRUE);
	}
    }
  }

  working = FALSE;
  return TRUE;
}

/* Button_release event handler of drawing_area. */
static gint
button_release_event (GtkWidget *widget, GdkEventButton *event)
{
  pressed = FALSE;
  return TRUE;
}

/* Motion_notify event handler of drawing_area. Draws/erases depending
 * on the value of drawing, and also sets the bitmap of current_glyph. */
static gint
motion_notify_event(GtkWidget *widget, GdkEventButton *event)
{
  int row, column;

  if (pressed == FALSE)
    return TRUE;
  if (current_glyph == NULL)
    return TRUE;
  if (working)
    return TRUE;
  if ((event->x >= pixmap_width) || (event->y >= pixmap_height))
    return TRUE;

  working = TRUE;

  row = event->y / (boxsize + 1);
  column = event->x / (boxsize + 1);

  switch (GFE_FONT_TYPE(current_font)) {
    case FONT_TYPE_RASTER:
      if (drawing) {
	gfe_bitmap_set_bit(GFE_RASTER_GLYPH(current_glyph)->bitmap,
			   column, row);
	color_box(column, row, TRUE);
      }
      else {
	gfe_bitmap_unset_bit(GFE_RASTER_GLYPH(current_glyph)->bitmap,
			column, row);
	color_box(column, row, FALSE);
      }
  }
 
  working = FALSE;
  return TRUE;
}

/* Draws the current editing glyph on the drawing area. */
static void
draw_glyph(void)
{
  static int w_old = 0, h_old = 0; 
  int r, c;
  int bbx_width, bbx_height;
  GdkRectangle update_rect;

  if (current_glyph == NULL){
    /* no font, screen should be cleared */
    if (pixmap) {
      gdk_draw_rectangle(pixmap,
		      drawing_area->style->bg_gc[GTK_WIDGET_STATE(drawing_area)],
		      TRUE, 0, 0, pixmap_width, pixmap_height);
      update_rect.x = 0;
      update_rect.y = 0;
      update_rect.width = w_old;
      update_rect.height = h_old;
      gtk_widget_draw(drawing_area, &update_rect);
      w_old = 0;
      h_old = 0;
      while (gtk_events_pending())
        gtk_main_iteration();
      gdk_pixmap_unref(pixmap);
      pixmap = NULL;
      gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 1, 1);
    }
    return;
  }

  switch (GFE_GLYPH_TYPE(current_glyph)) {
    case FONT_TYPE_RASTER:
      if (pixmap) gdk_pixmap_unref(pixmap);

      bbx_width = GFE_RASTER_GLYPH(current_glyph)->bitmap->width;
      bbx_height = GFE_RASTER_GLYPH(current_glyph)->bitmap->height;

      pixmap_width = bbx_width * (boxsize + 1) + 1;
      pixmap_height = bbx_height * (boxsize + 1) + 1;

      pixmap = gdk_pixmap_new(drawing_area->window,
			      pixmap_width, pixmap_height, -1);
      gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area),
			    pixmap_width, pixmap_height);
      gdk_draw_rectangle(pixmap, drawing_area->style->white_gc,
		         TRUE, 0, 0, pixmap_width, pixmap_height);
      /* change this to increase in steps of boxsize + 1 */
      for (c = 0; c <= bbx_width; c++){
	gdk_draw_line(pixmap, drawing_area->style->black_gc,
		      c * (boxsize + 1), 0, c * (boxsize + 1), pixmap_height);
      }
      for (r = 0; r <= bbx_height; r++){
	gdk_draw_line(pixmap, drawing_area->style->black_gc,
		      0, r * (boxsize + 1), pixmap_width, r * (boxsize + 1));
      }

      for (r = 0; r < bbx_height; r++){
        for (c = 0; c < bbx_width; c++){
          if (gfe_bitmap_get_bit(GFE_RASTER_GLYPH(current_glyph)->bitmap,
				 c, r)){
	    color_box(c, r, TRUE);
	  }
	} 
      }
      break;
  }

  update_rect.x = 0;
  update_rect.y = 0;

  /* replace this part with MAX macro */
  update_rect.width = MAX(w_old, pixmap_width);
  update_rect.height = MAX(h_old, pixmap_height);
  gtk_widget_draw(drawing_area, &update_rect);

  w_old = pixmap_height;
  h_old = pixmap_width;

  while (gtk_events_pending())
    gtk_main_iteration();
}

