/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>
#include <malloc.h>

#include "auxil.h"
#include "bdf.h"
#include "bitmap.h"
#include "editor.h"
#include "file.h"
#include "font.h"
#include "gui.h"
#include "raster_glyph.h"

/* Main window is deactivated during file open. */
extern GtkWidget *main_window;

/* Uses the font list */
extern GfeFont *first_font;
extern GfeFont *last_font;
extern GfeFont *current_font;
extern GfeGlyph *current_glyph;

/* local variables */

/* name of the last accessed file */
static char *last_file = NULL;

/* local function declarations */

static void file_open_ok(GtkWidget *w, GtkFileSelection *filesel);
static void file_save_as_ok(GtkWidget *w, GtkFileSelection *filesel);
static gboolean is_open(char *filename);
static void dialog_destroy(GtkWidget *widget, gpointer data);
static int font_format_from_extension(char *filename);

/* global functions */

/* File->open menu item handler. Desplays the file selection box and
 * attempts to open a file. If open is successful, an item is added
 * to the end of the font list, and current font is set to point to
 * that. Status bar and drawing area is also updated. Only tries
 * BDF format (now). */
void
file_open_callback(void)
{
  GtkWidget *filesel;
  int i;
  char tmp;

  gtk_widget_set_sensitive(main_window, FALSE);
  filesel = gtk_file_selection_new("Open file");
  gtk_signal_connect (GTK_OBJECT (filesel), "destroy",
		  (GtkSignalFunc) dialog_destroy,
		  (gpointer)filesel);
  gtk_signal_connect(
		  GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		  "clicked", (GtkSignalFunc) file_open_ok, filesel);
  gtk_signal_connect(
		  GTK_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button), 
		  "clicked", (GtkSignalFunc) dialog_destroy, 
		  (gpointer)filesel);

  /* set the last accessed directory */
  if (last_file) {
    i = strlen(last_file);
    while (i && (last_file[i] != '/'))
      i--;
    if (last_file[i] == '/'){
      tmp = last_file[i + 1];
      last_file[i + 1] = 0;
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), last_file);
      last_file[i + 1] = tmp;
    }
    else {
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), "");
    }
  }
  else {
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), "");
  }
  gtk_widget_show(filesel);
}

void
file_save_as_callback(void)
{
  GtkWidget *filesel;

  gtk_widget_set_sensitive(main_window, FALSE);
  filesel = gtk_file_selection_new("Save as");
  gtk_signal_connect (GTK_OBJECT (filesel), "destroy",
		  (GtkSignalFunc) dialog_destroy,
		  (gpointer)filesel);
  gtk_signal_connect(
		  GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		  "clicked", (GtkSignalFunc) file_save_as_ok, filesel);
  gtk_signal_connect(
		  GTK_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button), 
		  "clicked", (GtkSignalFunc) dialog_destroy, 
		  (gpointer)filesel);

  if (current_font) {
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel),
		                    current_font->filename);
  }
  else {
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), "");
  }

  gtk_widget_show(filesel);
}

/* File->save menu item handler. Currently saves only raster glyphs
 * in BDF format. */
void
file_save_callback(void)
{
  if (current_font == NULL) return;
  /* switch to save as if file name is not defined */
  if (current_font->filename == NULL)
    file_save_as_callback();
  set_current_glyph_pre();
  file_save(current_font);
}

/* File->close menu item handler. Destroy current_font, attempts to set
 * currnet_font to the previous item, next item or NULL. */
void
file_close_callback(void)
{
  GfeFont *font;

  if (current_font == NULL) return;

  set_current_glyph_pre();
  font = current_font;
  if (current_font->previous) {
    if ((current_font->previous->next = current_font->next)){
      current_font->next->previous = current_font->previous;
    }
    else {
      /* this is the last font */
      last_font = current_font->previous;
    } 
    current_font = current_font->previous;
  }
  else if (current_font->next) {
    /* this is the first font */
    current_font->next->previous = NULL;
    current_font = current_font->next;
    first_font = current_font;
  }
  else {
    /* this is the only font */
    first_font = NULL;
    last_font = NULL;
    current_font = NULL;
  }

  switch (GFE_FONT_TYPE(font)) {
    case FONT_TYPE_RASTER:
      gfe_raster_font_destroy(GFE_RASTER_FONT(font));
      break;
    default:
      g_print("Unknown font type in close\n");
  }

  set_current_glyph_post();

  set_title();
  if (first_font)
    status_bar_set_string(current_font->filename);
  else
    status_bar_set_string("GNU Font Editor");
}

/* Opens the file (currently) of a given format. If format is
 * FONT_FORMAT_NONE attempts to guess from the extension. If the
 * extansion is unknown and format is FONT_FORMAT_NONE, attempts to open
 * it using all known formats. If filename is NULL, just return
 * successfully (zero). file_open Returns zero on success, nonzero
 * on failure. */
int
file_open(char *filename, int format)
{
  GfeFont *font;

  if (filename == NULL) return 0;

  /* if format is not given, guess from extension */
  if (format == FONT_FORMAT_NONE)
    format = font_format_from_extension(filename);

  switch (format) {
    case FONT_FORMAT_BDF21:
    case FONT_FORMAT_BDF22:
      font = GFE_FONT(bdf_font_read_from_file(filename));
      break;
    default:
      /* can't figure out format. Try all known ones */
      font = GFE_FONT(bdf_font_read_from_file(filename));
  }

  if (font == NULL) return -1;

  /* add the opened font to the current font list */
  set_current_glyph_pre();
  if (first_font) {
    /* font list empty */
    last_font->next = font;
    font->previous = last_font;
    font->next = NULL;
    last_font = font;
    current_font = font;
  }
  else {
    /* add after the last font */
    first_font = font;
    last_font = font;
    current_font = font;
    font->next = NULL;
    font->previous = NULL;
  }
  set_current_glyph_post();

  set_title();
  status_bar_set_string(filename);
  string_copy(filename, &last_file);

  return 0;
}

/* Saves font to the file given by its filename field. Returns zero on
 * success, nonzero on error. */
int file_save(GfeFont *font)
{
  switch (font->format) {
    case FONT_FORMAT_BDF21:
    case FONT_FORMAT_BDF22:
      return (bdf_font_save(GFE_RASTER_FONT(current_font)));
    default:
      /* unknown format */
      g_print("Unknown font type in save\n");
      return -1;
  }
}

/* local functions */

/* Opens file if <ok> button is pressed in file selection dialog box. */
static void
file_open_ok(GtkWidget *widget, GtkFileSelection *filesel)
{
  char *filename;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION (filesel));
  gtk_widget_set_sensitive(GTK_WIDGET(filesel), FALSE);

  if (is_open(filename)){
    gtk_widget_destroy(GTK_WIDGET(filesel));
    gtk_widget_set_sensitive(main_window, TRUE);
    return;
  }

  if (!file_open(filename, 0)){
    gtk_widget_destroy(GTK_WIDGET(filesel));
    gtk_widget_set_sensitive(main_window, TRUE);
    return;
  }

  /* File open failed. A dialog box shoud appear here to ask for
   * a retry. Currently just wait for another filename. */
  gtk_widget_set_sensitive(GTK_WIDGET(filesel), TRUE);
}

static void
file_save_as_ok(GtkWidget *widget, GtkFileSelection *filesel)
{
  char *filename, *save_filename = NULL;
  int format, save_format;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION (filesel));
  gtk_widget_set_sensitive(GTK_WIDGET(filesel), FALSE);

  format = font_format_from_extension(filename);

  if (format != FONT_FORMAT_NONE) {
    save_format = current_font->format;
    current_font->format = format;
    string_copy(current_font->filename, &save_filename);
    gfe_font_set_filename(current_font, filename);

    set_current_glyph_pre();

    if (!file_save(current_font)){
      status_bar_set_string(current_font->filename);
      gtk_widget_destroy(GTK_WIDGET(filesel));
      gtk_widget_set_sensitive(main_window, TRUE);
      free(save_filename);
      set_title();
      return;
    }
    current_font->format = save_format;
    gfe_font_set_name(current_font, save_filename);
    free(save_filename);
  }

  /* save unsuccessful */
  gtk_widget_set_sensitive(GTK_WIDGET(filesel), TRUE);
  g_print("save unsuccessful\n");
}

/* Returns TRUE is the file given by absolute pathname filename is
 * already open, FALSE otherwise. */
static
gboolean is_open(char *filename)
{
  GfeFont *font;
  if (first_font == NULL) return FALSE;

  font = first_font;

  while (font) {
    if (strcmp(font->filename, filename) == 0) return TRUE;
    font = font->next;
  }

  return FALSE;
}

/* Destroy event handler of file selection box. */
static void
dialog_destroy(GtkWidget *widget, gpointer data)
{
  gtk_widget_set_sensitive(main_window, TRUE);
  gtk_widget_destroy(GTK_WIDGET(data));
}

/* returns the font format of a filename using it's extension. */
static int font_format_from_extension(char *filename)
{
  int i;

  i = strlen(filename);
  
  /* look for the rightmost '.' */
  while ((filename[i] != '.') && (i > 0))
    i--;

  /* no extension or dot file */
  if (i == 0)
    return FONT_FORMAT_NONE;

  i++;

  if ((strcmp(filename + i, "bdf") == 0) || 
      (strcmp(filename + i, "BDF") == 0))
    return FONT_FORMAT_BDF21;
  else
    return FONT_FORMAT_NONE;
}
