/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>
#include <malloc.h>

#include "auxil.h"
#include "font.h"
#include "glyph.h"

/* global functions */

/* Initializes a font that is already allocated. Should NOT be called
 * directly. Instead routines that create derived types from GfeFont
 * should call this just after creating a pointer to the derived type.
 * see structures.txt. */
void
gfe_font_init(GfeFont *font)
{
  font->format = FONT_FORMAT_NONE;
  font->name = NULL;
  font->filename = NULL;
  font->content_version = NULL;
  font->num_glyphs = 0;
  font->first_glyph = NULL;
  font->last_glyph = NULL;
  font->current_glyph = NULL;
  font->next = NULL;
  font->previous = NULL;
}

/* Sets the 8 least significant bits of the flags to type. */
void
gfe_font_set_type(GfeFont *font, guint8 type)
{
  font->flags &= 0xffffff00;
  font->flags |= type;
}

/* Sets the name of the font to name. Returns 0 on success, non 
 * zero on failure. */
int
gfe_font_set_name(GfeFont *font, char *name)
{
  return(string_copy(name, &(font->name)));
}

/* Sets the filename of the font to filename. Returns 0 on success,
 * non zero on failure. */
int
gfe_font_set_filename(GfeFont *font, char *filename)
{
  return(string_copy(filename, &(font->filename)));
}

/* Sets the content version of the font to content_version. Returns
 * 0 on success, non zero on failure. */
int
gfe_font_set_content_version(GfeFont *font, char *content_version)
{
  return(string_copy(content_version, &(font->content_version)));
}

/* Clears the contents of a font. See structures.txt */
void
gfe_font_free(GfeFont *font)
{
  free(font->name);
  free(font->filename);
  free(font->content_version);
  font->name = NULL;
  font->filename = NULL;
  font->content_version = NULL;
}

/* adds a glyph to the font after all the items with equal/smaller
 * values of encoding. Doesn't change the current_glyph of the font,
 * unless the glyph list is empty. If the glyph list is empty, it is
 * set to point to glyph. */
void
gfe_font_add_glyph(GfeFont *font, GfeGlyph *glyph)
{
  GfeGlyph *tmp;
  int done;

  if (font->first_glyph) {
    /* glyph list not empty */
    if (font->first_glyph->encoding > glyph->encoding){
      /* add before the fitst_glyph */
      glyph->next = font->first_glyph;
      font->first_glyph->previous = glyph;
      glyph->previous = NULL;
      font->first_glyph = glyph;
    }
    else {
      tmp = font->first_glyph;
      done = FALSE;
      while (!done) {
	if ((tmp = tmp->next)) {
	  if (tmp->encoding > glyph->encoding){
            /* add before tmp */
            glyph->next = tmp;
            glyph->previous = tmp->previous;
            tmp->previous->next = glyph;
            tmp->previous = glyph;
	    done = TRUE;
	  }
	}
	else {
          /* add after last_glyph */
	  font->last_glyph->next = glyph;
	  glyph->previous = font->last_glyph;
	  glyph->next = NULL;
	  font->last_glyph = glyph;
	  done = TRUE;
	}
      }
    }
  }
  else {
    /* glyph list empty */
    font->first_glyph = glyph;
    font->first_glyph->next = NULL;
    font->first_glyph->previous = NULL;
    font->last_glyph = glyph;
    font->current_glyph = glyph;
  }
  (font->num_glyphs)++;
}

