/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>
#include <malloc.h>

#include "auxil.h"
#include "character.h"
#include "editor.h"
#include "font.h"
#include "glyph_toolbar.h"
#include "gui.h"
#include "menu.h"
#include "toolbar.h"

/* Main window widget. */
GtkWidget *main_window;

/* Current font used by set_title() */
extern GfeFont *current_font;

/* Status bar and context_id */
static GtkWidget *statusbar;
static guint statusbar_context_id;

/* local function declarations */

static gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data);

/* Global functions */

/* Creates the GUI of GFE. Returns zero on success, nonzero on failure. */
int
create_gui(void)
{  
  GtkWidget *mainvbox;
  GtkWidget *hbox;
  GtkWidget *drawingarea;
  GtkWidget *scr_window;
  GtkWidget *menu_bar;
  GtkWidget *toolbar;
  GtkWidget *glyph_toolbar;

  /* main window */
  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_signal_connect(GTK_OBJECT(main_window), "delete_event",
		     GTK_SIGNAL_FUNC (delete_event), NULL);
  gtk_signal_connect(GTK_OBJECT(main_window), "destroy",
		     GTK_SIGNAL_FUNC(destroy_callback), NULL);

  /* window properties */
  set_title();
  gtk_container_set_border_width(GTK_CONTAINER(main_window), 2);
  gtk_window_set_policy(GTK_WINDOW(main_window), TRUE, TRUE, TRUE);
  gtk_widget_set_usize(main_window, 600, 450);
  gtk_widget_realize(main_window);

  /* main vertical box */
  mainvbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(main_window), mainvbox);
  gtk_widget_show(mainvbox);

  /* menu bar */
  menu_bar = create_menu_bar(main_window);
  gtk_box_pack_start(GTK_BOX(mainvbox), menu_bar, FALSE, FALSE, 0);
  gtk_widget_show(menu_bar);

   /* toolbar */
  toolbar = create_main_toolbar(main_window);
  gtk_box_pack_start(GTK_BOX(mainvbox), toolbar, FALSE, FALSE, 0);
  gtk_widget_show(toolbar);
  
  /* hbox to hold drawing area and glyph toolbar */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(mainvbox), hbox, TRUE, TRUE, 0);
  gtk_widget_show(hbox);

  /* scrolled window to hold drawing area */
  scr_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scr_window),
		                 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_AUTOMATIC);
  gtk_widget_set_usize(scr_window, 30, 20);
  gtk_box_pack_start(GTK_BOX(hbox), scr_window, TRUE, TRUE, 0);
  gtk_widget_show(scr_window);

  /* drawing area */
  drawingarea = create_drawing_area();
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scr_window),
		                        drawingarea);
  gtk_widget_show(drawingarea);

  /* glyph toolbar */
  glyph_toolbar = create_glyph_toolbar(main_window);
  gtk_box_pack_start(GTK_BOX(hbox), glyph_toolbar, FALSE, FALSE, 10);
  gtk_widget_show(glyph_toolbar);

  /* status bar */
  statusbar = gtk_statusbar_new();
  statusbar_context_id = gtk_statusbar_get_context_id(
		  GTK_STATUSBAR(statusbar), TITLE_STRING);
  gtk_box_pack_start(GTK_BOX(mainvbox), statusbar, FALSE, FALSE, 0);
  gtk_widget_show(statusbar);

  gtk_widget_show(main_window);

  status_bar_set_string("GNU Font Editor");

  return 0;
}

/* Sets the title of the main window to TITLE_STRING, plus if a
 * file is open, it's name (not the absolute name). */
void
set_title(void)
{
  char *title = NULL;
  gboolean have_filename = TRUE;
  int i, l;

  /* check for the availability of filename */
  if (current_font == NULL)
    have_filename = FALSE;
  else if (current_font->filename == NULL)
    have_filename = FALSE;

  if (have_filename) {
    i = l = strlen(current_font->filename);
    while ((i > 0) && (current_font->filename[i] != '/'))
      i--;
    if (current_font->filename[i] != '/')
      i--;
    
    title = malloc(sizeof(char)*strlen(TITLE_STRING) + 3 + l - i + 1);
    /* if setting the title is failed, just ignore */
    if (title == NULL)
      return;
    strcpy(title, TITLE_STRING);
    strcpy(title + strlen(TITLE_STRING), " - ");
    strcpy(title + strlen(TITLE_STRING) + 3, current_font->filename + i + 1);
  }
  else {
    string_copy(TITLE_STRING, &title);
  }

  gtk_window_set_title(GTK_WINDOW(main_window), title);

  free(title);
}

/* Pop's the previous string from the statusbar if it has any, and 
 * pushes status_string, if it is not NULL */
void
status_bar_set_string(char *status_string)
{
  static depth = 0;
  if (depth > 0) {
    /* status bar not empty */
    gtk_statusbar_pop(GTK_STATUSBAR(statusbar), statusbar_context_id);
    depth--;
  }
  if (status_string) {
    /* a valid string */
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), statusbar_context_id,
		       status_string);
    depth++;
  }
}

/* Destroy event handler of the main window. */
void
destroy_callback(GtkWidget *widget, gpointer data)
{
  while (gtk_events_pending())
    gtk_main_iteration();

  gtk_main_quit();
}

/* local functions */

/* Delete event handler of main_window. A confirmation dialog
 * box should be added here. */
static gint
delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  return(FALSE);
}

