/* GFE -- GNU Font Editor
 * Copyright (C) 2000 Anuradha Ratnaweera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>

#include <malloc.h>

#include "bitmap.h"
#include "font.h"
#include "glyph.h"
#include "raster_glyph.h"

/* global functions */

/* Creates a GfeRasterFont and returns a pointer to it. Returns
 * NULL in case of failure. See structures.txt. */
GfeRasterGlyph*
gfe_raster_glyph_new(GfeRasterFont *font)
{
  GfeRasterGlyph *glyph;

  glyph = malloc(sizeof(GfeRasterGlyph));

  gfe_glyph_init(GFE_GLYPH(glyph));

  gfe_glyph_set_type(GFE_GLYPH(glyph), FONT_TYPE_RASTER);

  glyph->sw_x = font->sw_x;
  glyph->sw_y = font->sw_y;
  glyph->dw_x = font->dw_x;
  glyph->dw_y = font->dw_y;
  glyph->sw1_x = font->sw1_x;
  glyph->sw1_y = font->sw1_y;
  glyph->dw1_x = font->dw1_x;
  glyph->dw1_y = font->dw1_y;
  glyph->vv_x = font->vv_x;
  glyph->vv_y = font->vv_y;
  glyph->bbx_width = font->bbx_width;
  glyph->bbx_height = font->bbx_height;
  glyph->bbx_x_offset = font->bbx_x_offset;
  glyph->bbx_y_offset = font->bbx_y_offset;

  /* this is an exception to the convention of setting all pointers
   * to NULL, as in structures.txt. */
  glyph->bitmap = gfe_bitmap_new(); 

  return (glyph);
}

/* Frees the dynamically allocated space of glyph. See structures.txt. */
void
gfe_raster_glyph_free(GfeRasterGlyph *glyph)
{
  gfe_glyph_free(GFE_GLYPH(glyph));
  gfe_bitmap_destroy(glyph->bitmap);
}

/* Frees the dynamically allocated space of glyph, and destroys it.
 * See structures.txt. */
void
gfe_raster_glyph_destroy(GfeRasterGlyph *glyph)
{
  gfe_raster_glyph_free(glyph);
  free(glyph);
}

/* Copies the glyph to font's current_glyph after prooning. The contents
 * of font->current_glyph will be cleared first. Returns zero on success,
 * non zero on failure. */
int
gfe_raster_glyph_copy_to_font(GfeRasterGlyph *glyph, GfeRasterFont *font)
{
  int r, c;
  int max_row = 0, max_col = 0;
  int min_row = glyph->bitmap->height;
  int min_col = glyph->bitmap->width;
  int found;

  for (r = 0; r < glyph->bitmap->height; r++){
    found = FALSE;
    for (c = 0; c < glyph->bitmap->width; c++){
      if (gfe_bitmap_get_bit(glyph->bitmap, c, r)){
	found = TRUE;
	if (min_col > c) min_col = c;
	if (max_col < c) max_col = c;
      }
    }
    if (found){
      if (min_row > r) min_row = r;
      if (max_row < r) max_row = r;
    }
  }

  /* blank glyph - add special code here */
  if (min_row > max_row) return 1;

  gfe_glyph_set_name(GFE_FONT(font)->current_glyph, GFE_GLYPH(glyph)->name);
  GFE_FONT(font)->current_glyph->encoding = GFE_GLYPH(glyph)->encoding;

  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw_x = glyph->sw_x;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw_y = glyph->sw_y;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw_x = glyph->dw_x;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw_y = glyph->dw_y;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw1_x = glyph->sw1_x;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw1_y = glyph->sw1_y;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw1_x = glyph->dw1_x;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw1_y = glyph->dw1_y;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->vv_x = glyph->vv_x;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->vv_y = glyph->vv_y;

  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_width
	  = max_col - min_col + 1;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_height
	  = max_row - min_row + 1;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_x_offset
	  = glyph->bbx_x_offset + min_col;
  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_y_offset
	  = glyph->bbx_y_offset + glyph->bitmap->height - 1 - max_row;

  gfe_bitmap_allocate(
		  GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bitmap,
		  max_col - min_col + 1, max_row - min_row + 1);

  for (r = min_row; r <= max_row; r++){
    for (c = min_col; c <= max_col; c++){
      if (gfe_bitmap_get_bit(glyph->bitmap, c, r)){
	gfe_bitmap_set_bit(
			GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bitmap,
			c - min_col, r - min_row);
      }
    }
  }

  return 0;
}

/* Copies the font's current_glyph to glyph. The widgh and height and 
 * origin(s) of glyph are not changed. The bitmap of glyph will first be
 * cleared. Returns zero on success, non zero on failure. */
int
gfe_raster_glyph_copy_from_font(GfeRasterFont *font, GfeRasterGlyph *glyph)
{
  int rows, cols;
  int row, col;
  int x_off, y_off;
  int x_shift, y_shift;

  if (glyph == NULL) return 1;

  rows = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_height;
  cols = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_width;
  x_off = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_x_offset;
  y_off = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bbx_y_offset;
  
  GFE_GLYPH(glyph)->encoding = GFE_FONT(font)->current_glyph->encoding;

  x_shift = x_off - glyph->bbx_x_offset;
  y_shift = - y_off + glyph->bbx_height - rows + glyph->bbx_y_offset;

  gfe_bitmap_clear(glyph->bitmap);

  for (row = 0; row < rows; row++){
    for (col= 0; col< cols; col++){
      if (gfe_bitmap_get_bit(GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->bitmap,
			      col, row)) {
	gfe_bitmap_set_bit(glyph->bitmap, col+ x_shift, row + y_shift);
      }
    }
  }

  gfe_glyph_set_name(GFE_GLYPH(glyph), GFE_FONT(font)->current_glyph->name);
  GFE_GLYPH(glyph)->encoding = GFE_FONT(font)->current_glyph->encoding;

  glyph->sw_x = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw_x;
  glyph->sw_y = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw_y;
  glyph->dw_x = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw_x;
  glyph->dw_y = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw_y;
  glyph->sw1_x = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw1_x;
  glyph->sw1_y = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->sw1_y;
  glyph->dw1_x = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw1_x;
  glyph->dw1_y = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->dw1_y;
  glyph->vv_x = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->vv_x;
  glyph->vv_y = GFE_RASTER_GLYPH(GFE_FONT(font)->current_glyph)->vv_y;

  return 0;
}

